/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Portions created by Jeremy Lea
 *   Copyright 1999 Jeremy Lea <reg@shale.csir.co.za>.  All rights reserved.
 *
 * The contents of this file are subject to either Mozilla Public License
 * (You may obtain a copy of the License at http://www.mozilla.org/MPL/) or
 * the BSD style license included below; you may not use this file except in
 * compliance with one of these licenses.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * Mozilla Public License for the specific language governing rights and
 * limitations under the License.
 * 
 * The Original Code is The Mozilla SVG Implementation.
 * 
 * The Initial Developer of the Original Code is Jeremy Lea
 * 
 * Contributor(s): 
 *
 */

#include "nsSVGPathSeg.h"

nsSVGPathSeg::nsSVGPathSeg()
{
	mType = kSVG_PATHCOORD_RELATIVE;
  mPrev = nsnull;
  mNext = nsnull;
}

nsSVGPathSeg::nsSVGPathSeg(nsSVGCoordType type, nsSVGPathSeg* segment)
{
	mType = type;
  
 	if (nsnull != segment)
 	{
 		mPrev = segment;
 		mNext = segment->mNext;
 		segment->mNext = this;
 		if (nsnull != mNext)
 			mNext->mPrev = this;
 	}
}

nsSVGPathSeg::~nsSVGPathSeg()
{
	if (nsnull != mNext)
		mNext->mPrev = mPrev;
	if (nsnull != mPrev)
		mPrev->mNext = mNext;
  mPrev = mNext = nsnull;
}

NS_METHOD_(PRUint32)
nsSVGPathSeg::pathLength()
{
	PRUint32 i = 1;
	nsSVGPathSeg * p = mNext;
  while (nsnull != p)
  {	
  	i++;
  	p = p->mNext;
  }
	return pathIndex() + i;
}
  
NS_METHOD_(PRUint32)
nsSVGPathSeg::pathIndex()
{
	PRUint32 i = 0;
	nsSVGPathSeg * p = mPrev;
  while (nsnull != p)
  {	
  	i++;
  	p = p->mPrev;
  }
	return i;
}

NS_METHOD
nsSVGPathSeg::appendSVGPathSeg(nsSVGPathSeg** segment)
{
  if (nsnull == segment)
  	return NS_ERROR_NULL_POINTER;
  
  if (nsnull == *segment)
  	return NS_ERROR_NULL_POINTER;

  nsSVGPathSeg* t = mNext;
  mNext = *segment;
  *segment = this;
  while (nsnull != (*segment)->mNext)
  	*segment = (*segment)->mNext;
  (*segment)->mNext = t;
  t->mPrev = *segment;

  return NS_OK;
}

NS_METHOD
nsSVGPathSeg::insertSVGPathSeg(nsSVGPathSeg** segment)
{
  if (nsnull == mPrev)
		return appendSVGPathSeg(segment);
	else
		return mPrev->appendSVGPathSeg(segment);
}

NS_METHOD
nsSVGPathSeg::replaceSVGPathSeg(nsSVGPathSeg** segment)
{
	if (nsnull == segment)
		return NS_ERROR_NULL_POINTER;

  nsSVGPathSeg* tPrev = (*segment)->mPrev;
  nsSVGPathSeg* tNext = (*segment)->mNext;
	(*segment)->mPrev = mPrev;
	(*segment)->mNext = mNext;
  mPrev = tPrev;
  mNext = tNext;
	*segment = this;

	return NS_OK;
}

NS_METHOD
nsSVGPathSeg::removeSVGPathSeg(nsSVGPathSeg** segment)
{
	if (nsnull == segment)
		return NS_ERROR_NULL_POINTER;

	mNext->mPrev = mPrev;
	mPrev->mNext = mNext;
  if (nsnull != *segment)
  {
		mPrev = (*segment);
		mNext = (*segment)->mNext;
		(*segment)->mNext = this;
		if (nsnull != mNext)
			mNext->mPrev = this;
  }
  else
  {
		mPrev = nsnull;
		mNext = nsnull;
  	*segment = this;
	}

	return NS_OK;
}

NS_IMETHODIMP_(nsPoint)
nsSVGPathSeg::getInitialPoint()
{
	if (nsnull != mPrev)
		return mPrev->getInitialPoint();
	else
	{
		nsPoint p;
		p.x = (nscoord)0.0;
		p.y = (nscoord)0.0;
		return p;
	}
}
	
// The Close Path element.

nsSVGClosePath::nsSVGClosePath()
	:nsSVGPathSeg()
{
}

nsSVGClosePath::nsSVGClosePath(nsSVGPathSeg* segment)
	:nsSVGPathSeg(kSVG_PATHCOORD_RELATIVE, segment)
{
}

nsSVGClosePath::~nsSVGClosePath()
{
}

NS_IMETHODIMP
nsSVGClosePath::getAsString(nsString** d)
{
  if (nsnull == d)
  	return NS_ERROR_NULL_POINTER;
  	
  if (nsnull == *d)
  	(*d) = new nsString();

  (*d)->Append("z ");

	return NS_OK;
}

NS_IMETHODIMP_(PRUint32)
nsSVGClosePath::normalizedLength()
{
	return 1;
}

NS_IMETHODIMP
nsSVGClosePath::normalizedPath(nsSVGPathSeg** segment)
{
  if (nsnull == segment)
  	return NS_ERROR_NULL_POINTER;
  	
  (*segment) = new nsSVGClosePath();

	return NS_OK;
}

NS_IMETHODIMP_(PRBool)
nsSVGClosePath::isTail()
{
	return PR_TRUE;
}

NS_IMETHODIMP_(nsPoint)
nsSVGClosePath::getCurrentPoint()
{
	return getInitialPoint();
}

NS_IMETHODIMP_(nsPoint)
nsSVGClosePath::getStartControlPoint()
{
	if (nsnull == mPrev)
		return getCurrentPoint();
	else
		return mPrev->getCurrentPoint();
}

NS_IMETHODIMP_(nsPoint)
nsSVGClosePath::getEndControlPoint()
{
	return getCurrentPoint();
}

NS_IMETHODIMP_(PRUint32)
nsSVGClosePath::pointLength()
{
	return 1;
}

NS_IMETHODIMP_(nsPoint*)
nsSVGClosePath::getPoints(nsPoint points[])
{
	points[0] = getCurrentPoint();
	return &points[1];
}


// The Move To element.

nsSVGMoveTo::nsSVGMoveTo()
	:nsSVGPathSeg()
{
	mPoint.x = (nscoord)0.0;
	mPoint.y = (nscoord)0.0;
}

nsSVGMoveTo::nsSVGMoveTo(nsSVGCoordType type, nscoord px, nscoord py, nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mPoint.x = px;
	mPoint.y = py;
}

nsSVGMoveTo::nsSVGMoveTo(nsSVGCoordType type, nsPoint point, nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mPoint = point;
}

nsSVGMoveTo::~nsSVGMoveTo()
{
}

NS_IMETHODIMP
nsSVGMoveTo::getAsString(nsString** d)
{
  if (nsnull == d)
  	return NS_ERROR_NULL_POINTER;
  	
  if (nsnull == *d)
  	(*d) = new nsString();

	switch (mType)
	{
		case kSVG_PATHCOORD_ABSOLUTE:
			(*d)->Append("M ");
			break;
		case kSVG_PATHCOORD_RELATIVE:
			(*d)->Append("m ");
			break;
		case kSVG_PATHCOORD_IMPLIED:
			(*d)->Append(" ");
			break;
	}
  (*d)->Append(mPoint.x);
  (*d)->Append(" ");
  (*d)->Append(mPoint.y);
  (*d)->Append(" ");

	return NS_OK;
}

NS_IMETHODIMP_(PRUint32)
nsSVGMoveTo::normalizedLength()
{
	return 1;
}

NS_IMETHODIMP
nsSVGMoveTo::normalizedPath(nsSVGPathSeg** segment)
{
  if (nsnull == segment)
  	return NS_ERROR_NULL_POINTER;
  	
  (*segment) = new nsSVGMoveTo(kSVG_PATHCOORD_ABSOLUTE, getCurrentPoint(), *segment);

	return NS_OK;
}

NS_IMETHODIMP_(PRBool)
nsSVGMoveTo::isHead()
{
	return PR_TRUE;
}
			
NS_IMETHODIMP_(nsPoint)
nsSVGMoveTo::getCurrentPoint()
{
  nsPoint p = mPoint;
	switch (getType())
	{
		case kSVG_PATHCOORD_RELATIVE:
			p += mPrev->getCurrentPoint();
			break;
		default:
			break;
	}

	return p;
}

NS_IMETHODIMP_(nsPoint)
nsSVGMoveTo::getInitialPoint()
{
	return getCurrentPoint();
}

NS_IMETHODIMP_(nsPoint)
nsSVGMoveTo::getStartControlPoint()
{
	if (nsnull == mPrev)
		return getCurrentPoint();
	else
		return mPrev->getCurrentPoint();
}

NS_IMETHODIMP_(nsPoint)
nsSVGMoveTo::getEndControlPoint()
{
	return getCurrentPoint();
}

NS_IMETHODIMP_(PRUint32)
nsSVGMoveTo::pointLength()
{
	return 1;
}

NS_IMETHODIMP_(nsPoint*)
nsSVGMoveTo::getPoints(nsPoint points[])
{
	points[0] = getCurrentPoint();
	return &points[1];
}

// The Line To element.

nsSVGLineTo::nsSVGLineTo()
	:nsSVGPathSeg()
{
	mPoint.x = (nscoord)0.0;
	mPoint.y = (nscoord)0.0;
}

nsSVGLineTo::nsSVGLineTo(nsSVGCoordType type, nscoord px, nscoord py, nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mPoint.x = px;
	mPoint.y = py;
}

nsSVGLineTo::nsSVGLineTo(nsSVGCoordType type, nsPoint point, nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mPoint = point;
}

nsSVGLineTo::~nsSVGLineTo()
{
}

NS_IMETHODIMP
nsSVGLineTo::getAsString(nsString** d)
{
  if (nsnull == d)
  	return NS_ERROR_NULL_POINTER;
  	
  if (nsnull == *d)
  	(*d) = new nsString();

	switch (mType)
	{
		case kSVG_PATHCOORD_ABSOLUTE:
			(*d)->Append("L ");
			break;
		case kSVG_PATHCOORD_RELATIVE:
			(*d)->Append("l ");
			break;
		case kSVG_PATHCOORD_IMPLIED:
			(*d)->Append(" ");
			break;
	}
  (*d)->Append(mPoint.x);
  (*d)->Append(" ");
  (*d)->Append(mPoint.y);
  (*d)->Append(" ");

	return NS_OK;
}

NS_IMETHODIMP_(PRUint32)
nsSVGLineTo::normalizedLength()
{
	return 1;
}

NS_IMETHODIMP
nsSVGLineTo::normalizedPath(nsSVGPathSeg** segment)
{
  if (nsnull == segment)
  	return NS_ERROR_NULL_POINTER;
  	
  (*segment) = new nsSVGLineTo(kSVG_PATHCOORD_ABSOLUTE,	getCurrentPoint(), *segment);

	return NS_OK;
}

NS_IMETHODIMP_(nsPoint)
nsSVGLineTo::getCurrentPoint()
{
  nsPoint p = mPoint;
	switch (getType())
	{
		case kSVG_PATHCOORD_RELATIVE:
			p += mPrev->getCurrentPoint();
			break;
		default:
			break;
	}

	return p;
}

NS_IMETHODIMP_(nsPoint)
nsSVGLineTo::getStartControlPoint()
{
	if (nsnull == mPrev)
		return getCurrentPoint();
	else
		return mPrev->getCurrentPoint();
}

NS_IMETHODIMP_(nsPoint)
nsSVGLineTo::getEndControlPoint()
{
	return getCurrentPoint();
}

NS_IMETHODIMP_(PRUint32)
nsSVGLineTo::pointLength()
{
	return 1;
}

NS_IMETHODIMP_(nsPoint*)
nsSVGLineTo::getPoints(nsPoint points[])
{
	points[0] = getCurrentPoint();
	return &points[1];
}

// The Horizontal Line To element.

nsSVGHorizontalTo::nsSVGHorizontalTo()
	:nsSVGPathSeg()
{
	mLength = (nscoord)0.0;
}

nsSVGHorizontalTo::nsSVGHorizontalTo(nsSVGCoordType type, nscoord l, nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mLength = l;
}

nsSVGHorizontalTo::~nsSVGHorizontalTo()
{
}

NS_IMETHODIMP
nsSVGHorizontalTo::getAsString(nsString** d)
{
  if (nsnull == d)
  	return NS_ERROR_NULL_POINTER;
  	
  if (nsnull == *d)
  	(*d) = new nsString();

	switch (mType)
	{
		case kSVG_PATHCOORD_ABSOLUTE:
			(*d)->Append("H ");
			break;
		case kSVG_PATHCOORD_RELATIVE:
			(*d)->Append("h ");
			break;
		case kSVG_PATHCOORD_IMPLIED:
			(*d)->Append(" ");
			break;
	}
  (*d)->Append(mLength);
  (*d)->Append(" ");

	return NS_OK;
}

NS_IMETHODIMP_(PRUint32)
nsSVGHorizontalTo::normalizedLength()
{
	return 1;
}

NS_IMETHODIMP
nsSVGHorizontalTo::normalizedPath(nsSVGPathSeg** segment)
{
  if (nsnull == segment)
  	return NS_ERROR_NULL_POINTER;
  	
  (*segment) = new nsSVGLineTo(kSVG_PATHCOORD_ABSOLUTE, getCurrentPoint(), *segment);

	return NS_OK;
}

NS_IMETHODIMP_(nsPoint)
nsSVGHorizontalTo::getCurrentPoint()
{
  nsPoint p = mPrev->getCurrentPoint();
	switch (getType())
	{
		case kSVG_PATHCOORD_RELATIVE:
			p.x += mLength;
			break;
		default:
			p.x = mLength;
			break;
	}

	return p;
}

NS_IMETHODIMP_(nsPoint)
nsSVGHorizontalTo::getStartControlPoint()
{
	if (nsnull == mPrev)
		return getCurrentPoint();
	else
		return mPrev->getCurrentPoint();
}

NS_IMETHODIMP_(nsPoint)
nsSVGHorizontalTo::getEndControlPoint()
{
	return getCurrentPoint();
}

NS_IMETHODIMP_(PRUint32)
nsSVGHorizontalTo::pointLength()
{
	return 1;
}

NS_IMETHODIMP_(nsPoint*)
nsSVGHorizontalTo::getPoints(nsPoint points[])
{
	points[0] = getCurrentPoint();
	return &points[1];
}

// The Vertical Line To element.

nsSVGVerticalTo::nsSVGVerticalTo()
	:nsSVGPathSeg()
{
	mLength = (nscoord)0.0;
}

nsSVGVerticalTo::nsSVGVerticalTo(nsSVGCoordType type, nscoord l, nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mLength = l;
}

nsSVGVerticalTo::~nsSVGVerticalTo()
{
}

NS_IMETHODIMP
nsSVGVerticalTo::getAsString(nsString** d)
{
  if (nsnull == d)
  	return NS_ERROR_NULL_POINTER;
  	
  if (nsnull == *d)
  	(*d) = new nsString();

	switch (mType)
	{
		case kSVG_PATHCOORD_ABSOLUTE:
			(*d)->Append("V ");
			break;
		case kSVG_PATHCOORD_RELATIVE:
			(*d)->Append("v ");
			break;
		case kSVG_PATHCOORD_IMPLIED:
			(*d)->Append(" ");
			break;
	}
  (*d)->Append(mLength);
  (*d)->Append(" ");

	return NS_OK;
}

NS_IMETHODIMP_(PRUint32)
nsSVGVerticalTo::normalizedLength()
{
	return 1;
}

NS_IMETHODIMP
nsSVGVerticalTo::normalizedPath(nsSVGPathSeg** segment)
{
  if (nsnull == segment)
  	return NS_ERROR_NULL_POINTER;
  	
  (*segment) = new nsSVGLineTo(kSVG_PATHCOORD_ABSOLUTE, getCurrentPoint(), *segment);

	return NS_OK;
}

NS_IMETHODIMP_(nsPoint)
nsSVGVerticalTo::getCurrentPoint()
{
  nsPoint p = mPrev->getCurrentPoint();
	switch (getType())
	{
		case kSVG_PATHCOORD_RELATIVE:
			p.y += mLength;
			break;
		default:
			p.y = mLength;
			break;
	}

	return p;
}

NS_IMETHODIMP_(nsPoint)
nsSVGVerticalTo::getStartControlPoint()
{
	if (nsnull == mPrev)
		return getCurrentPoint();
	else
		return mPrev->getCurrentPoint();
}

NS_IMETHODIMP_(nsPoint)
nsSVGVerticalTo::getEndControlPoint()
{
	return getCurrentPoint();
}

NS_IMETHODIMP_(PRUint32)
nsSVGVerticalTo::pointLength()
{
	return 1;
}

NS_IMETHODIMP_(nsPoint*)
nsSVGVerticalTo::getPoints(nsPoint points[])
{
	points[0] = getCurrentPoint();
	return &points[1];
}

// The Curve To element.

nsSVGCubicCurveTo::nsSVGCubicCurveTo()
	:nsSVGPathSeg()
{
	mPoint.x = (nscoord)0.0;
	mPoint.y = (nscoord)0.0;
	mC1.x = (nscoord)0.0;
	mC1.y = (nscoord)0.0;
	mC2.x = (nscoord)0.0;
	mC2.y = (nscoord)0.0;
}

nsSVGCubicCurveTo::nsSVGCubicCurveTo(nsSVGCoordType type, nscoord x1, nscoord y1,
										nscoord x2, nscoord y2, nscoord px, nscoord py,
										nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mPoint.x = px;
	mPoint.y = py;
	mC1.x = x1;
	mC1.y = y1;
	mC2.x = x2;
	mC2.y = y2;
}

nsSVGCubicCurveTo::nsSVGCubicCurveTo(nsSVGCoordType type, nsPoint c1, nsPoint c2,
										nsPoint point, nsSVGPathSeg* segment)
	:nsSVGPathSeg(type, segment)
{
	mPoint = point;
	mC1 = c1;
	mC2 = c2;
}

nsSVGCubicCurveTo::~nsSVGCubicCurveTo()
{
}

NS_IMETHODIMP
nsSVGCubicCurveTo::getAsString(nsString** d)
{
  if (nsnull == d)
  	return NS_ERROR_NULL_POINTER;
  	
  if (nsnull == *d)
  	(*d) = new nsString();

	switch (mType)
	{
		case kSVG_PATHCOORD_ABSOLUTE:
			(*d)->Append("C ");
			break;
		case kSVG_PATHCOORD_RELATIVE:
			(*d)->Append("c ");
			break;
		case kSVG_PATHCOORD_IMPLIED:
			(*d)->Append(" ");
			break;
	}
  (*d)->Append(mC1.x);
  (*d)->Append(" ");
  (*d)->Append(mC1.y);
  (*d)->Append(" ");
  (*d)->Append(mC2.x);
  (*d)->Append(" ");
  (*d)->Append(mC2.y);
  (*d)->Append(" ");
  (*d)->Append(mPoint.x);
  (*d)->Append(" ");
  (*d)->Append(mPoint.y);
  (*d)->Append(" ");

	return NS_OK;
}

NS_IMETHODIMP_(PRUint32)
nsSVGCubicCurveTo::normalizedLength()
{
	return 1;
}

NS_IMETHODIMP
nsSVGCubicCurveTo::normalizedPath(nsSVGPathSeg** segment)
{
  if (nsnull == segment)
  	return NS_ERROR_NULL_POINTER;
  	
  (*segment) = new nsSVGCubicCurveTo(kSVG_PATHCOORD_ABSOLUTE,	getStartControlPoint(),
								  								getEndControlPoint(), getCurrentPoint(), *segment);

	return NS_OK;
}

NS_IMETHODIMP_(nsPoint)
nsSVGCubicCurveTo::getCurrentPoint()
{
  nsPoint p = mPoint;
	switch (getType())
	{
		case kSVG_PATHCOORD_RELATIVE:
			p += mPrev->getCurrentPoint();
			break;
		default:
			break;
	}

	return p;
}

NS_IMETHODIMP_(nsPoint)
nsSVGCubicCurveTo::getStartControlPoint()
{
  nsPoint p = mC1;
	switch (getType())
	{
		case kSVG_PATHCOORD_RELATIVE:
			p += mPrev->getCurrentPoint();
			break;
		default:
			break;
	}

	return p;
}

NS_IMETHODIMP_(nsPoint)
nsSVGCubicCurveTo::getEndControlPoint()
{
  nsPoint p = mC2;
	switch (getType())
	{
		case kSVG_PATHCOORD_RELATIVE:
			p += mPrev->getCurrentPoint();
			break;
		default:
			break;
	}

	return p;
}

NS_IMETHODIMP_(PRUint32)
nsSVGCubicCurveTo::pointLength()
{
	return 3;
}

NS_IMETHODIMP_(nsPoint*)
nsSVGCubicCurveTo::getPoints(nsPoint points[])
{
	points[0] = getStartControlPoint();
	points[1] = getEndControlPoint();
	points[2] = getCurrentPoint();
	return &points[3];
}

nsSVGVerticalTo t;
