/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Portions created by Jeremy Lea
 *   Copyright 1999 Jeremy Lea <reg@shale.csir.co.za>.  All rights reserved.
 *
 * The contents of this file are subject to either Mozilla Public License
 * (You may obtain a copy of the License at http://www.mozilla.org/MPL/) or
 * the BSD style license included below; you may not use this file except in
 * compliance with one of these licenses.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * Mozilla Public License for the specific language governing rights and
 * limitations under the License.
 * 
 * The Original Code is The Mozilla SVG Implementation.
 * 
 * The Initial Developer of the Original Code is Jeremy Lea
 * 
 * Contributor(s): 
 *
 */

#include "prtypes.h"
#include "nsIContent.h"
#include "nsIAtom.h"
#include "nsIPresContext.h"
#include "nsIStyleContext.h"
#include "nsINameSpaceManager.h"

#include "nsString.h"
#include "nsCSSRendering.h"
#include "nsFrame.h"

#include "nsSVGFrames.h"
#include "nsSVGAtoms.h"

// This file contains definitions of path segments for an SVG <path>. The
// base class contains the basic operations, along with code for a
// supporting linked list implementation.  The other classes are built on
// this, one for each type of path element.

// WARNING:  This class is a hack.  It looks like a doubly linked list but
// it's not.  It's two doubly linked lists on top of one another.  The
// second list is of subpaths, which are defined by movetos and closepaths. 
// It also has a self extending head, so you don't want these on a stack or
// as a data member. You wont be able to get at it's head or tail.  Just
// hold on to one segment, and it will do the rest.  Don't hold on to
// another because it might get deleted out from under you...
//
// Please don't play with this code unless you've had several cups of
// coffee...

// These are the various path segment types.
enum nsSVGPathSegType {
	kSVG_PATHSEG_UNKNOWN                      = 0,  // ?
	kSVG_PATHSEG_CLOSEPATH                    = 1,  // z
	kSVG_PATHSEG_MOVETO_ABS                   = 2,  // M
	kSVG_PATHSEG_MOVETO_REL                   = 3,  // m
	kSVG_PATHSEG_LINETO_ABS                   = 4,  // L
	kSVG_PATHSEG_LINETO_REL                   = 5,  // l
	kSVG_PATHSEG_CURVETO_CUBIC_ABS            = 6,  // C
	kSVG_PATHSEG_CURVETO_CUBIC_REL            = 7,  // c
	kSVG_PATHSEG_CURVETO_QUADRATIC_ABS        = 8,  // Q
	kSVG_PATHSEG_CURVETO_QUADRATIC_REL        = 9,  // q
	kSVG_PATHSEG_ARC_SWEEP_ABS                = 10, // A
	kSVG_PATHSEG_ARC_SWEEP_REL                = 11, // a
	kSVG_PATHSEG_ARC_VECTOR_ABS               = 12, // B
	kSVG_PATHSEG_ARC_VECTOR_REL               = 13, // b
	kSVG_PATHSEG_LINETO_HORIZONTAL_ABS        = 14, // H
	kSVG_PATHSEG_LINETO_HORIZONTAL_REL        = 15, // h
	kSVG_PATHSEG_LINETO_VERTICAL_ABS          = 16, // V
	kSVG_PATHSEG_LINETO_VERTICAL_REL          = 17, // v
	kSVG_PATHSEG_CURVETO_CUBIC_SMOOTH_ABS     = 18, // S
	kSVG_PATHSEG_CURVETO_CUBIC_SMOOTH_REL     = 19, // s
	kSVG_PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS = 20, // T
	kSVG_PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL = 21  // t
};

enum nsSVGCoordType
{
	kSVG_PATHCOORD_ABSOLUTE,
	kSVG_PATHCOORD_RELATIVE,
	kSVG_PATHCOORD_IMPLIED
};

class nsSVGPathSeg
{
public:
  // default constructor.
	nsSVGPathSeg();
	
	// Appending contructor.  Used to append new elements to (*segment).
	nsSVGPathSeg(nsSVGCoordType type, nsSVGPathSeg* segment);

  // default destructor.  Removes the entire list, in both directions.
	virtual ~nsSVGPathSeg();

	// The following methods are designed to support the SVG DOM implmentation.

	// The length of this path.
	NS_METHOD_(PRUint32) pathLength();
  
	// The index of this segment on the path.  Zero based.
	NS_METHOD_(PRUint32) pathIndex();

	// Return a string containing the minimal representation of this path
	// segment.  If (*d) is not nsnull the data is appended.
	NS_IMETHOD getAsString(nsString** d) = 0;

	// Appends a path segment(s) after this segment (inserting or
	// extending the list).  Returns the last element inserted.  Must be able
	// to be called recursively.  This class assumes ownership of the appended
	// segment(s), so don't delete them.
	NS_METHOD appendSVGPathSeg(nsSVGPathSeg** segment);

	// Inserts a path segment(s) before this segment.  If this is called on
	// the first element in a path, then the data is effectively appended. 
	// Same rules as above.
	NS_METHOD insertSVGPathSeg(nsSVGPathSeg** segment);

	// Replace this segment.  The old segment (ie this) is returned.  If the
	// segment is part of a list (ie prev or next != NULL), then the returned
	// segment is automatically added into that list (ie the two segments are
	// swapped).  Keep another copy of segment else you you'll leak the path.
	// XXX: need better semantics.
	NS_METHOD replaceSVGPathSeg(nsSVGPathSeg** segment);

	// Remove this segment.  The removed segment is returned and should be
	// deleted.  You better have another hold on the list or you'll leak it.
	// XXX: need better semantics.
	NS_METHOD removeSVGPathSeg(nsSVGPathSeg** segment);

	// Return the number of segments needed to create a normalized version of
	// this segment of the path.  See SVG spec.
	NS_IMETHOD_(PRUint32) normalizedLength() = 0;

	// Return a normalized representation of this path.  The returned path
	// must be deleted.
	NS_IMETHOD normalizedPath(nsSVGPathSeg** segment) = 0;

	// The following methods are designed to help paint paths.

	// The type of path segment.
	NS_IMETHOD_(nsSVGPathSegType) getSVGPathSegType() {
		return kSVG_PATHSEG_UNKNOWN;
	};

	// The type of coordinates used.
	NS_IMETHOD_(nsSVGCoordType) getType() {
		if (nsnull == mPrev)
			return kSVG_PATHCOORD_ABSOLUTE;
		else
			return mType;
	};

	// Is this the start or end of a subpath?  Default provided.
	// These are NOT the start and end of the linked list.
	NS_IMETHOD_(PRBool) isHead() {
		if (nsnull == mPrev)
			return PR_TRUE;
		else
			return PR_FALSE;
	};
	NS_IMETHOD_(PRBool) isTail() {
		if (nsnull == mNext)
			return PR_TRUE;
		else
			return PR_FALSE;
	};
  NS_METHOD_(nsSVGPathSeg*) getHead() {
  	nsSVGPathSeg* p = this;
  	while (PR_FALSE == p->isHead())
  		p = p->mPrev;
  	return p;
  };
  NS_METHOD_(nsSVGPathSeg*) getTail() {
  	nsSVGPathSeg* p = this;
  	while (PR_FALSE == p-isTail())
  		p = p->mNext;
  	return p;
  };
	
	// The 'current point' as set by this segment.
  NS_IMETHOD_(nsPoint) getCurrentPoint() = 0;

  // The 'initial point' as seen by this segment.  Default provided.
	NS_IMETHOD_(nsPoint) getInitialPoint();

  // The 'control points' as set by this segment.
  NS_IMETHOD_(nsPoint) getStartControlPoint() = 0;
  NS_IMETHOD_(nsPoint) getEndControlPoint() = 0;

  // The number of vertices need to represent this segment.
  NS_IMETHOD_(PRUint32) pointLength() = 0;

  // The vertices need to represent this segment.  Excludes the initial
  // vertex.  The array passed in must be big enough to at least hold
  // pointLength() points.  The pointer returned is points[pointLength()].
  NS_IMETHOD_(nsPoint*) getPoints(nsPoint points[]) = 0;
  
protected:

	nsSVGCoordType	mType;
	nsSVGPathSeg*		mPrev;
	nsSVGPathSeg*		mNext;
};

// The close path element.  Does not have to be the last PathSeg in a list.

class nsSVGClosePath : public nsSVGPathSeg
{
public:
	nsSVGClosePath();
	nsSVGClosePath(nsSVGPathSeg* segment);
	virtual ~nsSVGClosePath();

	NS_IMETHOD getAsString(nsString** d);

	NS_IMETHOD_(PRUint32) normalizedLength();
	NS_IMETHOD normalizedPath(nsSVGPathSeg** segment);

	NS_IMETHOD_(nsSVGPathSegType) getSVGPathSegType() {
		return kSVG_PATHSEG_CLOSEPATH;
	};

	NS_IMETHOD_(PRBool) isTail();
  NS_IMETHOD_(nsPoint) getCurrentPoint();
  NS_IMETHOD_(nsPoint) getStartControlPoint();
  NS_IMETHOD_(nsPoint) getEndControlPoint();

  NS_IMETHOD_(PRUint32) pointLength();
  NS_IMETHOD_(nsPoint*) getPoints(nsPoint points[]);
  
protected:

};

class nsSVGMoveTo : public nsSVGPathSeg
{
public:
	nsSVGMoveTo();
	nsSVGMoveTo(nsSVGCoordType type, nscoord px, nscoord py, nsSVGPathSeg* segment);
	nsSVGMoveTo(nsSVGCoordType type, nsPoint point, nsSVGPathSeg* segment);
	virtual ~nsSVGMoveTo();

	NS_IMETHOD getAsString(nsString** d);

	NS_IMETHOD_(PRUint32) normalizedLength();
	NS_IMETHOD normalizedPath(nsSVGPathSeg** segment);

	NS_IMETHOD_(nsSVGPathSegType) getSVGPathSegType() {
		if (kSVG_PATHCOORD_RELATIVE == getType())
			return kSVG_PATHSEG_MOVETO_REL;
		else
			return kSVG_PATHSEG_MOVETO_ABS;
	};

	NS_IMETHOD_(PRBool) isHead();
  NS_IMETHOD_(nsPoint) getCurrentPoint();
	NS_IMETHOD_(nsPoint) getInitialPoint();
  NS_IMETHOD_(nsPoint) getStartControlPoint();
  NS_IMETHOD_(nsPoint) getEndControlPoint();

  NS_IMETHOD_(PRUint32) pointLength();
  NS_IMETHOD_(nsPoint*) getPoints(nsPoint points[]);
  
protected:

	nsPoint mPoint;
};

class nsSVGLineTo : public nsSVGPathSeg
{
public:
	nsSVGLineTo();
	nsSVGLineTo(nsSVGCoordType type, nscoord px, nscoord py, nsSVGPathSeg* segment);
	nsSVGLineTo(nsSVGCoordType type, nsPoint point, nsSVGPathSeg* segment);
	virtual ~nsSVGLineTo();

	NS_IMETHOD getAsString(nsString** d);

	NS_IMETHOD_(PRUint32) normalizedLength();
	NS_IMETHOD normalizedPath(nsSVGPathSeg** segment);

	NS_IMETHOD_(nsSVGPathSegType) getSVGPathSegType() {
		if (kSVG_PATHCOORD_RELATIVE == getType())
			return kSVG_PATHSEG_LINETO_REL;
		else
			return kSVG_PATHSEG_LINETO_ABS;
	};

  NS_IMETHOD_(nsPoint) getCurrentPoint();
  NS_IMETHOD_(nsPoint) getStartControlPoint();
  NS_IMETHOD_(nsPoint) getEndControlPoint();

  NS_IMETHOD_(PRUint32) pointLength();
  NS_IMETHOD_(nsPoint*) getPoints(nsPoint points[]);
  
protected:

	nsPoint mPoint;
};

class nsSVGHorizontalTo : public nsSVGPathSeg
{
public:
	nsSVGHorizontalTo();
	nsSVGHorizontalTo(nsSVGCoordType type, nscoord l, nsSVGPathSeg* segment);
	virtual ~nsSVGHorizontalTo();

	NS_IMETHOD getAsString(nsString** d);

	NS_IMETHOD_(PRUint32) normalizedLength();
	NS_IMETHOD normalizedPath(nsSVGPathSeg** segment);

	NS_IMETHOD_(nsSVGPathSegType) getSVGPathSegType() {
		if (kSVG_PATHCOORD_RELATIVE == getType())
			return kSVG_PATHSEG_LINETO_REL;
		else
			return kSVG_PATHSEG_LINETO_ABS;
	};

  NS_IMETHOD_(nsPoint) getCurrentPoint();
  NS_IMETHOD_(nsPoint) getStartControlPoint();
  NS_IMETHOD_(nsPoint) getEndControlPoint();

  NS_IMETHOD_(PRUint32) pointLength();
  NS_IMETHOD_(nsPoint*) getPoints(nsPoint points[]);
  
protected:

	nscoord mLength;
};

class nsSVGVerticalTo : public nsSVGPathSeg
{
public:
	nsSVGVerticalTo();
	nsSVGVerticalTo(nsSVGCoordType type, nscoord l, nsSVGPathSeg* segment);
	virtual ~nsSVGVerticalTo();

	NS_IMETHOD getAsString(nsString** d);

	NS_IMETHOD_(PRUint32) normalizedLength();
	NS_IMETHOD normalizedPath(nsSVGPathSeg** segment);

	NS_IMETHOD_(nsSVGPathSegType) getSVGPathSegType() {
		if (kSVG_PATHCOORD_RELATIVE == getType())
			return kSVG_PATHSEG_LINETO_REL;
		else
			return kSVG_PATHSEG_LINETO_ABS;
	};

  NS_IMETHOD_(nsPoint) getCurrentPoint();
  NS_IMETHOD_(nsPoint) getStartControlPoint();
  NS_IMETHOD_(nsPoint) getEndControlPoint();

  NS_IMETHOD_(PRUint32) pointLength();
  NS_IMETHOD_(nsPoint*) getPoints(nsPoint points[]);
  
protected:

	nscoord mLength;
};

class nsSVGCubicCurveTo : public nsSVGPathSeg
{
public:
	nsSVGCubicCurveTo();
	nsSVGCubicCurveTo(nsSVGCoordType type, nscoord x1, nscoord y1,
										nscoord x2, nscoord y2, nscoord px, nscoord py,
										nsSVGPathSeg* segment);
	nsSVGCubicCurveTo(nsSVGCoordType type, nsPoint c1, nsPoint c2,
										nsPoint point, nsSVGPathSeg* segment);
	virtual ~nsSVGCubicCurveTo();

	NS_IMETHOD getAsString(nsString** d);

	NS_IMETHOD_(PRUint32) normalizedLength();
	NS_IMETHOD normalizedPath(nsSVGPathSeg** segment);

	NS_IMETHOD_(nsSVGPathSegType) getSVGPathSegType() {
		if (kSVG_PATHCOORD_RELATIVE == getType())
			return kSVG_PATHSEG_LINETO_REL;
		else
			return kSVG_PATHSEG_LINETO_ABS;
	};

  NS_IMETHOD_(nsPoint) getCurrentPoint();
  NS_IMETHOD_(nsPoint) getStartControlPoint();
  NS_IMETHOD_(nsPoint) getEndControlPoint();

  NS_IMETHOD_(PRUint32) pointLength();
  NS_IMETHOD_(nsPoint*) getPoints(nsPoint points[]);
  
protected:

	nsPoint mPoint;
	nsPoint mC1;
	nsPoint mC2;
};

	// Parse a string, appending path segments, returning the last segment on
	// the path.  The string must conform to the specifications from the SVG
	// spec, else an error is reported.  (*tail) contains the last element in
	// the list, and (*tail) should == nsnull on input else it will leak. 
	// Must be able to be called recursively.
	// appendPathData(nsString* d, nsSVGPathSeg** tail);
